/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) GREYC - IMAGE
 *
 * GREYC- IMAGE
 * 6 Boulevard Marchal Juin
 * F-14050 Caen Cedex France
 *
 *
 * This file is part of the PANTHEON Project
 * http://www.greyc.ensicaen.fr/~regis/Pantheon
 *
 *
 * This file is free software and is provided "AS IS" without a warranty
 * of any kind. Redistribution and use of this source code, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * - THE ENTIRE RISK TO THE QUALITY AND PERFORMANCE OF THIS SOFTWARE IS
 *   WITH YOU.
 * - THE GREYC SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE
 *   AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS FILE OR ITS
 *   DERIVATIVES.
 * - THE NAME OF THE GREYC MAY NOT BE USED TO ENDORSE OR PROMOTE PRODUCTS
 *   DERIVED FROM THIS SOFTWARE WITHOUT SPECIFIC PRIOR WRITTEN PERMISSION.
 */

package fr.greyc.pantheon.thunder.model;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Iterator;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.ImageIcon;
import javax.swing.JOptionPane;

import fr.greyc.pantheon.thunder.controller.Thunder;
import fr.greyc.pantheon.thunder.main.Main;

/**
 * Manages the available languages.
 * 
 * @author Julian CHENARD - 2009-04-07 
 * @author Jrmy GAUTIER - 2009-04-07
 * @author Rgis CLOUARD - 2009-10-06
 */
public class Language {
    /** Absolute path to the user's configuration file. */
    private final static String LANGUAGE_PROPERTY = Main.CONFIG_FOLDER + Main.CONFIG_FILE;

    private final static String DEFAULT_LANGUAGE = "en";
    
    /**
     * Returns the language specified in the configuration file if any.
     * The default language is English: "en".
     * @return the current language.
     */
    public String getLanguage( ) {
        String lang;
        String currentLang = DEFAULT_LANGUAGE;
 
        if ((lang = readLanguage()) != null) {
            currentLang = lang;
        } else {
            /* Get the current language of the user's session.*/
            String defaultLanguage = System.getProperty("user.language");
            /* If the language is not recognized in the application */
            /* the use English language. */
            if (Thunder.SUPPORTED_LOCALES.containsKey(defaultLanguage)) {
                currentLang = defaultLanguage;
            }
        }
        return currentLang;
    }

    /**
     * Read the preferred language in the configuration file.
     * @return the language string or null if no configuration file.
     */
    private String readLanguage( ) {
        File language_conf = new File(LANGUAGE_PROPERTY);
        if (!language_conf.exists()) {
            return null;
        }
        BufferedReader input;
        try {
            input = new BufferedReader(new FileReader(language_conf));
            String language = input.readLine();
            input.close();
            return language;
        } catch (FileNotFoundException ex) {
            System.out.println("Erreur d'ouverture du fichier " + language_conf.getAbsolutePath());
            Logger.getLogger(Language.class.getName()).log(Level.SEVERE, null, ex);
        } catch (IOException ex) {
            System.out.println("Erreur de lecture dans le fichier " + language_conf.getAbsolutePath());
            Logger.getLogger(Language.class.getName()).log(Level.SEVERE, null, ex);
        }

        return null;
    }

    /**
     * Writes the preferred language in the configuration file.
     * @param preferedLanguage the preferred language.
     */
    private void writeLanguageInFile( String preferedLanguage ) {
        File language_folder = new File(Main.CONFIG_FOLDER);
        try {
            if (!language_folder.exists() && !language_folder.mkdir()) {
                JOptionPane.showMessageDialog(null,
                		Translator.displayValue(Main.currentLocale, "language_warning1")+                		
                        language_folder.getAbsolutePath() +
                		Translator.displayValue(Main.currentLocale, "language_warning2"),
                		Translator.displayValue(Main.currentLocale, "language_warning_title"),
                		JOptionPane.WARNING_MESSAGE);
                return;
            }
        } catch (SecurityException ex) {
            Logger.getLogger(Language.class.getName()).log(Level.SEVERE, null, ex);
        }

        File language_conf = new File(LANGUAGE_PROPERTY);

        BufferedWriter input;
        try {
            language_conf.createNewFile();
            input = new BufferedWriter(new FileWriter(language_conf));
            input.write(preferedLanguage);
            input.close();

        } catch (FileNotFoundException ex) {
            System.out.println("Erreur d'ouverture du fichier " + language_conf.getAbsolutePath());
            Logger.getLogger(Language.class.getName()).log(Level.SEVERE, null, ex);
        } catch (IOException ex) {
            System.out.println("Erreur de lecture dans le fichier " + language_conf.getAbsolutePath());
            Logger.getLogger(Language.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /**
     * Prompts the user to choose the preferred language.
     * The default language is English.
     */
    private String askLanguage( ) {
        String[] available = getSupportedLocale();
        
        String s = (String) JOptionPane.showInputDialog(
                null,
                Translator.displayValue(Main.currentLocale, "chooselanguage"),
                Translator.displayValue(Main.currentLocale, "languagetitle"),
                JOptionPane.PLAIN_MESSAGE,
                new ImageIcon(getClass().getClassLoader().getResource(Main.iconDir+"information_16.png")),
                available,
                Main.currentLocale.getLanguage());

        if ((s != null) && (s.length() > 0)) {
            return s;
        }
        return DEFAULT_LANGUAGE;
    }

    /**
     * Changes and saves the current language.
     * @return the preferred language.
     */
    public String changeLanguage( ) {
        String clang = askLanguage();
        writeLanguageInFile(clang);
        return clang;
    }

    /**
     * Returns the list of available languages.
     * @return the list of languages names.
     */
    private String[] getSupportedLocale( ) {
        String[] available = new String[Thunder.SUPPORTED_LOCALES.size()];
        Iterator<String> it = Thunder.SUPPORTED_LOCALES.keySet().iterator();
        for (int i = 0; i < available.length; i++) {
            available[i] = it.next();
        }
        return available;
    }
}