/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) GREYC - IMAGE
 *
 * GREYC- IMAGE
 * 6 Boulevard Marchal Juin
 * F-14050 Caen Cedex France
 *
 *
 * This file is part of the PANTHEON Project
 * http://www.greyc.ensicaen.fr/~regis/Pantheon
 *
 *
 * This file is free software and is provided "AS IS" without a warranty
 * of any kind. Redistribution and use of this source code, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * - THE ENTIRE RISK TO THE QUALITY AND PERFORMANCE OF THIS SOFTWARE IS
 *   WITH YOU.
 * - THE GREYC SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE
 *   AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS FILE OR ITS
 *   DERIVATIVES.
 * - THE NAME OF THE GREYC MAY NOT BE USED TO ENDORSE OR PROMOTE PRODUCTS
 *   DERIVED FROM THIS SOFTWARE WITHOUT SPECIFIC PRIOR WRITTEN PERMISSION.
 */

package fr.greyc.pantheon.thunder.model.image;

import java.awt.AWTException;
import java.awt.image.ColorModel;
import java.awt.image.ImageConsumer;
import java.awt.image.ImageProducer;
import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Random;

/**
 * Extends ImageProducer to decode Pandore images. This code is based
 * on the book "Java AWT references" by John Zukowski.
 *
 * Custom usage:
 *       PandoreImageDecoder pan= new PandoreImageDecoder();
 *       pan.readImage(imagebyte);
 *       Image img = Toolkit.getDefaultToolkit().createImage(pan);
 *
 * @author Rgis Clouard - 2001-01-03
 * @version Rgis Clouard - 2002-11-07 (Pandore images).
 * @version Rgis Clouard - 2003-01-16 (Pandore 4).
 * @version Rgis Clouard - 2009-04-01 (add 3d images).
 */
public class PandoreImageDecoder implements ImageProducer {
	/* List of the Pandore types. */
	/* Do not change the values, they are conventional. */

	//	private static final int Po_Collection = 1;
//	private static final int Po_Img1duc = 2;
//	private static final int Po_Img1dsl = 3;
//	private static final int Po_Img1dsf = 4;

	private static final int Po_Img2duc = 5;
	private static final int Po_Img2dsl = 6;
	private static final int Po_Img2dsf = 7;

	private static final int Po_Img3duc = 8;
	private static final int Po_Img3dsl = 9;
	private static final int Po_Img3dsf = 10;

//	private static final int Po_Reg1d = 11;
	private static final int Po_Reg2d = 12;
	private static final int Po_Reg3d = 13;

//	private static final int Po_Graph2d = 14;
//	private static final int Po_Graph3d = 15;

	private static final int Po_Imc2duc = 16;
	private static final int Po_Imc2dsl = 17;
	private static final int Po_Imc2dsf = 18;

	private static final int Po_Imc3duc = 19;
	private static final int Po_Imc3dsl = 20;
	private static final int Po_Imc3dsf = 21;

//	private static final int Po_Imx1duc = 22;
//	private static final int Po_Imx1dsl = 23;
//	private static final int Po_Imx1dsf = 24;
//	private static final int Po_Imx2duc = 25;
//	private static final int Po_Imx2dsl = 26;
//	private static final int Po_Imx2dsf = 27;
//	private static final int Po_Imx3duc = 28;
//	private static final int Po_Imx3dsl = 29;
//	private static final int Po_Imx3dsf = 30;

	private static final String PANDOREMAGIC = "PANDORE04";

	////////// Header attributes //////////
	/** Magic number : PO_MAGIC */
	private String _magic = null;
	/** object type */
	private int _potype = 0;
	private boolean _reverseByteOrder = false;
	/** Since done in-memory; only one consumer */
	private ImageConsumer _consumer;

	/** Problem with big-endian and little-endian: direct reading or reverse byte order? */
	private boolean _loadError = false;

//	/** image: number of bands */
//	private int bands;

	/** Image width */
	private int _width=0;
	/** Image height */
	private int _height=0;
	/** Image depth */
	private int _depth=0;

	/** Number of  regions */
	private int _nlabels=0;

	/** Array of pixels. */
	private int _imageData[][];

	/** Format of Pandore file is single pass/frame, w/ complete scan lines in order. */
	private static int PandoreHints = (
			ImageConsumer.TOPDOWNLEFTRIGHT |
			ImageConsumer.COMPLETESCANLINES |
			ImageConsumer.SINGLEPASS |
			ImageConsumer.SINGLEFRAME
	);

	/**
	 * Reads Pandore file header and checks consistency.
	 * @param dis	buffered image data
	 * @return the identifier of the Pandore object if 'dis' is a regular Pandore
	 *         file or -1 otherwise.
	 * @exception AWTException	when file is not a regular Pandore file.
	 * @exception IOException	in case of io exception.
	 */
	private int readHeader( DataInputStream dis ) throws AWTException,	IOException {
		byte[] buf = new byte[10];

		// 1. read the magic number.
		dis.read(buf, 0, 9);
		_magic = new String(buf, 0, 9);
		dis.skipBytes(1); // skip '\n'
		if (!(PANDOREMAGIC.equals(_magic)))
			throw (new AWTException("Invalid Pandore Format " + _magic));

		// 2. skip unused bytes.
		dis.skipBytes(2);

		// 3. trick: When typeoj is too high -> reverse byte order.
		_potype = dis.readInt();
		if (_potype > 65535) {
			_reverseByteOrder = true;
			_potype = swap(_potype);
		}

		// 3. read creator name.
		dis.read(buf, 0, 8);
//		String ident = new String(buf, 0, 8);
		dis.skipBytes(1); // skip "\n"

		// 4. read creation date.
		dis.read(buf, 0, 9);
//		String date = new String(buf, 0, 9);
		dis.skipBytes(1); // '\n';

		// Skip unused bytes.
		dis.skipBytes(1);

		return _potype;
	}

	/**
	 * There is only a single consumer. When it registers, produce image. On
	 * error, notify consumer
	 * @param ic new consumer to be added.
	 */
	@Override
	public synchronized void addConsumer( ImageConsumer ic ) {
		_consumer = ic;
		try {
			produce();
		} catch (Exception e) {
			if (_consumer != null)
				_consumer.imageComplete(ImageConsumer.IMAGEERROR);
		}
		_consumer = null;
	}

	/**
	 * If consumer passed to routine is single consumer, return true, else false
	 * @param ic	consumer.
	 */
	@Override
	public synchronized boolean isConsumer( ImageConsumer ic ) {
		return (ic == _consumer);
	}

	/**
	 * Disables consumer if currently consuming
	 * @param ic	consumer.
	 */
	@Override
	public synchronized void removeConsumer( ImageConsumer ic ) {
		if (_consumer == ic)
			_consumer = null;
	}

	/**
	 * Production is done by adding consumer
	 * @param ic	consumer.
	 */
	@Override
	public void startProduction( ImageConsumer ic ) {
		addConsumer(ic);
	}

	/**
	 * Registers the specified ImageConsumer object as a consumer and starts an immediate
	 * reconstruction of the image data which will then be delivered to this consumer
	 * and any other consumer which might have already been registered with the producer.
	 * @param ic the specified ImageConsumer.
	 */
	@Override
	public void requestTopDownLeftRightResend( ImageConsumer ic ) {
		// Not needed. The data is always in this format.
	}

	/**
	 * Production Process: Prerequisite: Image already read into store array.
	 * (readImage) props / width / height already set (readImage) Assumes RGB
	 * Color Model - would need to filter to change. Sends Pandore Image data to
	 * consumer. Pixels sent one row at a time.
	 */
	private void produce( ) {
		ColorModel cm = ColorModel.getRGBdefault();
		if (_consumer != null) {
			if (_loadError) {
				_consumer.imageComplete(ImageConsumer.IMAGEERROR);
			} else {
				_consumer.setDimensions(_width, _height);
				_consumer.setColorModel(cm);
				_consumer.setHints(PandoreHints);
				for (int y = 0; y < _height; y++)
					_consumer.setPixels(0, y, _width, 1, cm, _imageData[y], 0, _width);
				_consumer.imageComplete(ImageConsumer.STATICIMAGEDONE);
			}
		}
	}

	/**
	 * Swaps an integer (from big to little endian or conversely).
	 * @param i	the integer to be swapped.
	 * @return the swapped integer.
	 */
	private  int swap( int i ) {
		int byte0 = i & 0xff;
		int byte1 = (i >> 8) & 0xff;
		int byte2 = (i >> 16) & 0xff;
		int byte3 = (i >> 24) & 0xff;
		// swap the byte order
		return (byte0 << 24) | (byte1 << 16) | (byte2 << 8) | byte3;
	}

	/**
	 * Reads an integer format the input stream.
	 * If needed inverts the bytes.
	 * @param dis the input stream.
	 * @return the data array.
	 * @throws IOException
	 */
	private int readInt( DataInputStream dis ) throws IOException {
		byte[] row = new byte[4];
		for (int b=0; b<4; b++)
			row[b]=dis.readByte();
		return parseBytes(0, row,4);
	}

	/**
	 * Converts a Pandore pixel to a color that can be used by Java.
	 * @param x			column number of the pixel.
	 * @param pixel		a Pandore pixel.
	 * @param bytes	bytes per Pandore pixel.
	 * @return the color to be displayed.
	 */
	private int parseBytes( int x, byte[] pixel, int bytes ) {
		int[] colorByte = new int[4];

		for (int b = 0; b < bytes; b++) {
			colorByte[b] = (int)pixel[(x*bytes) + b];
			// Problem signed -> unsigned.
			if (colorByte[b] < 0)
				colorByte[b] += 256;
		}
		int color = 0;

		if (_reverseByteOrder) {
			for (int b = 0; b < bytes; b++) {
				int shift = b * 8;
				color |= (colorByte[b] << shift);
			}
		} else {
			for (int b = 0; b < bytes; b++) {
				int shift = (bytes-1-b) * 8;
				color |= (colorByte[b] << shift);
			}
		}

		return color;
	}

	/**
	 * Reads color data image.
	 * @param dis buffered data to parse.
	 * @return the data array.
	 * @throws IOException
	 */
	private  int[][] readColorByteData( DataInputStream dis ) throws IOException {
		int[][] store = new int[_height][_width];
		byte row[];
		row = new byte[_width];
		for (int y = 0; y < _height; y++) {
			dis.readFully(row);
			for (int x = 0; x < _width; x++) {
				int color = parseBytes(x,row,1);
				store[y][x] = (0xff << 24) | (color << 16);
			}
		}
		dis.skipBytes(_height*_depth*_width);
		for (int y = 0; y < _height; y++) {
			dis.readFully(row);
			for (int x = 0; x < _width; x++) {
				int color = parseBytes(x, row, 1);
				store[y][x] |= (color << 8);
			}
		}
		dis.skipBytes(_height*_depth*_width);
		for (int y = 0; y < _height; y++) {
			dis.readFully(row);
			for (int x = 0; x < _width; x++) {
				int color = parseBytes(x, row, 1);
				store[y][x] |= color;
			}
		}

		return store;
	}

	/**
	 * Reads gray data image.
	 * @param dis buffered data to parse.
	 * @return the data array.
	 * @throws IOException
	 */
	private  int[][] readColorLongData( DataInputStream dis ) throws IOException {
		int[][] store = new int[_height][_width];

		int[][][] data = new int[3][_height][_width];
		int minVal=Integer.MAX_VALUE;
		int maxVal=Integer.MIN_VALUE;
		int datum;

		for (int b=0; b<3; b++) {
			// Read values.
			for (int y = 0; y < _height; y++) {
				for (int x = 0; x < _width; x++) {
					datum = dis.readInt();
					if (minVal < datum) minVal=datum;
					if (maxVal < datum) maxVal=datum;
					data[b][y][x] =datum;
				}
			}
			dis.skipBytes((Integer.SIZE/8)*_height*_depth*_width);
		}

		// calculate slope (for all bands).
		Float slope;
		if (minVal>=0) {
			minVal=0;
			if (maxVal <=255)
				slope=1.0F;
			else
				slope =255.0F/maxVal;
		} else {
			if (maxVal==minVal)
				slope=1.0F;
			else
				slope=255.0F/(maxVal-minVal);
		}

		// linear transform
		for (int y = 0; y < _height; y++) {
			for (int x = 0; x < _width; x++) {
				int pixel = (int)(slope*(data[0][y][x]-minVal));
				store[y][x] = (0xff << 24) | (pixel << 16);
				pixel  = (int)(slope*(data[1][y][x]-minVal));
				store[y][x] |= (pixel << 8);
				pixel  = (int)(slope*(data[2][y][x]-minVal));
				store[y][x] |= pixel ;
			}
		}
		return store;
	}

	/**
	 * Reads color data image.
	 * @param dis buffered data to parse.
	 * @return the data array.
	 * @throws IOException
	 */
	private  int[][] readColorFloatData( DataInputStream dis ) throws IOException {
		int[][] store = new int[_height][_width];

		float[][][] data = new float[3][_height][_width];
		float minVal=Float.MAX_VALUE;
		float maxVal=Float.MIN_VALUE;
		float datum;

		for (int b=0; b<3; b++) {
			// Read values.
			for (int y = 0; y < _height; y++) {
				for (int x = 0; x < _width; x++) {
					datum = dis.readFloat();
					if (minVal < datum) minVal=datum;
					if (maxVal < datum) maxVal=datum;
					data[b][y][x] =datum;
				}
			}
			dis.skipBytes((Float.SIZE/8)*_height*_depth*_width);
		}

		// calculate slope (for all bands).
		Float slope;
		if (minVal>=0) {
			minVal=0;
			if (maxVal <=255.0F)
				slope=1.0F;
			else
				slope =255.0F/maxVal;
		} else {
			if (maxVal==minVal)
				slope=1.0F;
			else
				slope=255.0F/(maxVal-minVal);
		}

		// linear transform
		for (int y = 0; y < _height; y++) {
			for (int x = 0; x < _width; x++) {
				int pixel = (int)(slope*(data[0][y][x]-minVal));
				store[y][x] = (0xff << 24) | (pixel << 16);
				pixel  = (int)(slope*(data[1][y][x]-minVal));
				store[y][x] |= (pixel << 8);
				pixel  = (int)(slope*(data[2][y][x]-minVal));
				store[y][x] |= pixel ;
			}
		}
		return store;
	}

	/**
	 * Reads the specific header part of 2D color images.
	 * @param dis the input stream.
	 * @throws IOException in case of io errors.
	 */
	private  void read2DColorAttributes( DataInputStream dis ) throws IOException {
		readInt(dis); // bands = 3;
		_height = readInt(dis);
		_width = readInt(dis);

		// ColorSpace.
		dis.readInt();
	}

	/**
	 * Reads the specific header part of 3D color images.
	 * @param dis the input stream.
	 * @throws IOException in case of io errors.
	 */
	private  void read3DColorAttributes( DataInputStream dis ) throws IOException {
		readInt(dis); // bands = 3;
		_depth = readInt(dis);
		_height = readInt(dis);
		_width = readInt(dis);

		// ColorSpace.
		dis.readInt();
	}

	/**
	 * Reads the specific header part of 2D gray images.
	 * @param dis the input stream.
	 * @throws IOException in case of io errors.
	 */
	private  void read2DGrayAttributes( DataInputStream dis ) throws IOException {
		readInt(dis); // bands = 1 !
		_height = readInt(dis);
		_width = readInt(dis);
	}

	/**
	 * Reads the specific header part of 3D gray images.
	 * @param dis the input stream.
	 * @throws IOException in case of io errors.
	 */
	private  void read3DGrayAttributes( DataInputStream dis ) throws IOException {
		readInt(dis); // bands = 1 !
		_depth = readInt(dis);
		_height = readInt(dis);
		_width = readInt(dis);
	}

	/**
	 * Reads gray data image.
	 * @param dis buffered data to parse.
	 * @return the data array.
	 * @throws IOException
	 */
	private  int[][] readGrayByteData( DataInputStream dis ) throws IOException {
		int[][] store = new int[_height][_width];
		byte row[];

		row = new byte[_width];
		for (int y = 0; y < _height; y++) {
			dis.readFully(row);
			for (int x = 0; x < _width; x++) {
				int color = parseBytes(x, row, 1);
				store[y][x] = (0xff << 24) | (color << 16);
				store[y][x] |= (color << 8);
				store[y][x] |= color;
			}
		}
		return store;
	}

	/**
	 * Reads gray data image.
	 * @param dis buffered data to parse.
	 * @return the data array.
	 * @throws IOException
	 */
	private  int[][] readGrayLongData( DataInputStream dis ) throws IOException {
		int[][] store = new int[_height][_width];

		int[][] data = new int[_height][_width];
		int minVal=Integer.MAX_VALUE;
		int maxVal=Integer.MIN_VALUE;
		int datum;
		// Read values.
		for (int y = 0; y < _height; y++) {
			for (int x = 0; x < _width; x++) {
				datum = dis.readInt();
				if (minVal < datum) minVal=datum;
				if (maxVal < datum) maxVal=datum;
				data[y][x] =datum;
			}
		}

		// calculate slope (for all bands).
		Float slope;
		if (minVal>=0) {
			minVal=0;
			if (maxVal <=255)
				slope=1.0F;
			else
				slope =255.0F/maxVal;
		} else {
			if (maxVal==minVal)
				slope=1.0F;
			else
				slope=255.0F/(maxVal-minVal);
		}

		// linear transform
		for (int y = 0; y < _height; y++) {
			for (int x = 0; x < _width; x++) {
				datum = (int)(slope*(data[y][x]-minVal));
				store[y][x] = (0xff << 24) | (datum<< 16);
				store[y][x] |= (datum<< 8);
				store[y][x] |= datum;
			}
		}
		return store;
	}

	/**
	 * Reads gray data image.
	 * @param dis buffered data to parse.
	 * @return the data array.
	 * @throws IOException
	 */
	private  int[][] readGrayFloatData( DataInputStream dis ) throws IOException {
		int[][] store = new int[_height][_width];

		float[][] data = new float[_height][_width];
		float minVal=Integer.MAX_VALUE;
		float maxVal=Integer.MIN_VALUE;
		float datum;
		// Read values.
		for (int y = 0; y < _height; y++) {
			for (int x = 0; x < _width; x++) {
				datum = dis.readFloat();
				if (minVal < datum) minVal=datum;
				if (maxVal < datum) maxVal=datum;
				data[y][x] =datum;
			}
		}

		// Calculate slope (for all bands).
		Float slope;
		if (minVal>=0) {
			minVal=0;
			if (maxVal <=255)
				slope=1.0F;
			else
				slope =255.0F/maxVal;
		} else {
			if (maxVal==minVal)
				slope=1.0F;
			else
				slope=255.0F/(maxVal-minVal);
		}

		// linear transform.
		for (int y = 0; y < _height; y++) {
			for (int x = 0; x < _width; x++) {
				int pixel = (int)(slope*(data[y][x]-minVal));
				store[y][x] = (0xff << 24) | (pixel<< 16);
				store[y][x] |= (pixel<< 8);
				store[y][x] |= pixel;
			}
		}
		return store;
	}

	/**
	 * Reads the specific header part of 3D regions.
	 * @param dis the input stream.
	 * @throws IOException in case of io errors.
	 */
	private  void read3DRegionAttributes( DataInputStream dis ) throws IOException {
		readInt(dis); // bands = 3;
		_depth = readInt(dis);
		_height = readInt(dis);
		_width = readInt(dis);

		// number of labels.
		_nlabels = readInt(dis);
	}

	/**
	 * Reads the specific header part of 2D regions.
	 * @param dis the input stream.
	 * @throws IOException in case of io errors.
	 */
	private  void read2DRegionAttributes( DataInputStream dis ) throws IOException {
		readInt(dis); // bands = 3;
		_height = readInt(dis);
		_width = readInt(dis);

		// number of labels.
		_nlabels = readInt(dis);
	}

	/**
	 * Reads gray data image. First of all read the image size.
	 * @param dis buffered data to parse.
	 * @return the data array.
	 * @throws IOException
	 */
	private  int[][] readRegionData( DataInputStream dis ) throws IOException {
		int[][] store = new int[_height][_width];
		byte row[];

		Random rand = new Random(0); // Always the same seed !
		int[] lut = new int[_nlabels];
		int bytes;
		if (_nlabels < 256)
			bytes = 1;
		else if (_nlabels < 65536)
			bytes = 2;
		else
			bytes = 4;

		// Color are set randomly.
		for (int i = 0; i <_nlabels; i++)
			lut[i] = (0xff << 24) | (rand.nextInt(256) << 16)
			| (rand.nextInt(256) << 8) | rand.nextInt(256);

		row = new byte[_width * bytes];
		for (int y = 0; y < _height; y++) {
			dis.readFully(row);
			for (int x = 0; x < _width; x++) {
				int color = parseBytes(x, row, bytes);
				if (color<_nlabels)
					store[y][x] = lut[color];
			}
		}
		return store;
	}

	/**
	 * Reads image data from a stream.
	 * Parses data for Pandore format.
	 * Closes input stream when done.
	 * @param is an input stream.
	 */
	public void readImage( InputStream is ) {
		DataInputStream dis = null;
		BufferedInputStream bis = null;

		try {
			// Try a supported image.
			bis = new BufferedInputStream(is);
			dis = new DataInputStream(bis);

			readHeader(dis);
			switch (_potype) {
			case Po_Img2duc:
				read2DGrayAttributes(dis);
				_imageData = readGrayByteData(dis);
				break;
			case Po_Img2dsl:
				read2DGrayAttributes(dis);
				_imageData = readGrayLongData(dis);
				break;
			case Po_Img2dsf:
				read2DGrayAttributes(dis);
				_imageData = readGrayFloatData(dis);
				break;
			case Po_Reg2d:
				read2DRegionAttributes(dis);
				_imageData = readRegionData(dis);
				break;

			case Po_Imc2duc:
				read2DColorAttributes(dis);
				_imageData = readColorByteData(dis);
				break;
			case Po_Imc2dsl:
				read2DColorAttributes(dis);
				_imageData = readColorLongData(dis);
				break;
			case Po_Imc2dsf:
				read2DColorAttributes(dis);
				_imageData = readColorFloatData(dis);
				break;

				// For 3D we use the middle plane.
			case Po_Img3duc:
				read3DGrayAttributes(dis);
				dis.skipBytes((Byte.SIZE/8)*_height*_width*(_depth/2));
				_imageData = readGrayByteData(dis);
				break;
			case Po_Img3dsl:
				read3DGrayAttributes(dis);
				dis.skipBytes((Integer.SIZE/8)*_height*_width*(_depth/2));
				_imageData = readGrayLongData(dis);
				break;
			case Po_Img3dsf:
				read3DGrayAttributes(dis);
				dis.skipBytes((Float.SIZE/8)*_height*_width*(_depth/2));
				_imageData = readGrayFloatData(dis);
				break;
			case Po_Reg3d:
				read3DRegionAttributes(dis);
				_imageData = readRegionData(dis);
				break;

				// For 3D we use the middle plane.
			case Po_Imc3duc:
				read3DColorAttributes(dis);
				dis.skipBytes((Byte.SIZE/8)*_height*_width*(_depth/2));
				_imageData = readColorByteData(dis);
				break;
			case Po_Imc3dsl:
				read3DColorAttributes(dis);
				dis.skipBytes((Integer.SIZE/8)*_height*_width*(_depth/2));
				_imageData = readColorLongData(dis);
				break;
			case Po_Imc3dsf:
				read3DColorAttributes(dis);
				dis.skipBytes((Float.SIZE/8)*_height*_width*(_depth/2));
				_imageData = readColorFloatData(dis);
				break;
			}
		} catch (Exception io) {
			_loadError = true;
		} finally {
			try {
				if (dis != null) dis.close();
				if (bis != null) bis.close();
				if (is != null) is.close();
			} catch (IOException io) {
			}
		}
	}

	/**
	 * Reads the image content from a byte array.
	 * @param b a byte array.
	 */
	public void readImage( byte b[] ) {
		readImage(new ByteArrayInputStream(b));
	}

	/**
	 * Reads the image content from a file.
	 * @param filename a regular image file name.
	 */
	public void readImage( String filename ) {
		try {
			FileInputStream fileImage = new FileInputStream(filename);
			if (fileImage.available() > 0) {
				readImage(fileImage);
			}
			fileImage.close();
		} catch (IOException e) {
		}
	}
}